/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "Transmission/Thrift/ServiceSDKHandler.h"

#include "MixCast.h"
#include "Data/SdkData.h"
#include "MixCastLog.h"
#include "thrift/transport/TPipe.h"

#include "Async/Async.h"
#include "Engine/Engine.h"

void FServiceSDKHandler::SetActivationState(const bool active, const int64_t clientProcId)
{
	AsyncTask(ENamedThreads::GameThread, [=]()
	{
		if (!GEngine)
			return;

		if (IMixCast::IsAvailable())
			((FMixCast*)&IMixCast::Get())->SetActive(active, clientProcId);
	});
}

void FServiceSDKHandler::NotifyServiceStarted()
{
	ServiceStartedEvent.Broadcast();
}

void FServiceSDKHandler::UpdateCameraMetadata(const std::vector<VirtualCamera>& cameras)
{
	AsyncTask(ENamedThreads::GameThread, [=]()
	{
			std::vector<VirtualCamera> newList;

		for (int i = 0; i < cameras.size(); i++)
			{
				VirtualCamera* prevData = nullptr;
				if (cameras[i].__isset.identifier)
					prevData = SdkData::GetCameraById(cameras[i].identifier);
				if (prevData == nullptr && i < SdkData::VirtualCameras.size())
					prevData = &SdkData::VirtualCameras[i];

				if (!cameras[i].__isset.identifier && prevData == nullptr)
				{
					MIXCAST_ERROR("Incomplete video input list!");
					continue;
				}

				VirtualCamera newData = cameras[i];
				TransferCameraData(*prevData, newData);
				newList.push_back(newData);
			}

			SdkData::VirtualCameras.clear();
			for (int i = 0; i < newList.size(); i++)
				SdkData::VirtualCameras.push_back(newList[i]);
	});
}
void FServiceSDKHandler::TransferCameraData(const VirtualCamera& prevData, VirtualCamera& newData)
{
	if (!newData.__isset.identifier && prevData.__isset.identifier) {
		newData.identifier = prevData.identifier;
		newData.__isset.identifier = true;
	}

	if (!newData.__isset.fieldOfView && prevData.__isset.fieldOfView) {
		newData.fieldOfView = prevData.fieldOfView;
		newData.__isset.fieldOfView = true;
	}
	if (!newData.__isset.currentPosition && prevData.__isset.currentPosition) {
		newData.currentPosition = prevData.currentPosition;
		newData.__isset.currentPosition = true;
	}
	if (!newData.__isset.currentRotation && prevData.__isset.currentRotation) {
		newData.currentRotation = prevData.currentRotation;
		newData.__isset.currentRotation = true;
	}
	if (!newData.__isset.renderResolutionWidth && prevData.__isset.renderResolutionWidth) {
		newData.renderResolutionWidth = prevData.renderResolutionWidth;
		newData.__isset.renderResolutionWidth = true;
	}
	if (!newData.__isset.renderResolutionHeight && prevData.__isset.renderResolutionHeight) {
		newData.renderResolutionHeight = prevData.renderResolutionHeight;
		newData.__isset.renderResolutionHeight = true;
	}
	if (!newData.__isset.renderFramerate && prevData.__isset.renderFramerate) {
		newData.renderFramerate = prevData.renderFramerate;
		newData.__isset.renderFramerate = true;
	}
	if (!newData.__isset.usesForeground && prevData.__isset.usesForeground) {
		newData.usesForeground = prevData.usesForeground;
		newData.__isset.usesForeground = true;
	}
	if (!newData.__isset.usesFullRender && prevData.__isset.usesFullRender) {
		newData.usesFullRender = prevData.usesFullRender;
		newData.__isset.usesFullRender = true;
	}
	if (!newData.__isset.isFirstPerson && prevData.__isset.isFirstPerson) {
		newData.isFirstPerson = prevData.isFirstPerson;
		newData.__isset.isFirstPerson = true;
	}
	if (!newData.__isset.isBackgroundTranslucent && prevData.__isset.isBackgroundTranslucent) {
		newData.isBackgroundTranslucent = prevData.isBackgroundTranslucent;
		newData.__isset.isBackgroundTranslucent = true;
	}
	if (!newData.__isset.autoSnapshotEnabled && prevData.__isset.autoSnapshotEnabled) {
		newData.autoSnapshotEnabled = prevData.autoSnapshotEnabled;
		newData.__isset.autoSnapshotEnabled = true;
	}
	if (!newData.__isset.videoRecordingEnabled && prevData.__isset.videoRecordingEnabled) {
		newData.videoRecordingEnabled = prevData.videoRecordingEnabled;
		newData.__isset.videoRecordingEnabled = true;
	}
	if (!newData.__isset.videoStreamingEnabled && prevData.__isset.videoStreamingEnabled) {
		newData.videoStreamingEnabled = prevData.videoStreamingEnabled;
		newData.__isset.videoStreamingEnabled = true;
	}
}

void FServiceSDKHandler::UpdateVideoInputMetadata(const std::vector<VideoInput>& videoInputs)
{
	AsyncTask(ENamedThreads::GameThread, [=]()
		{
			std::vector<VideoInput> newList;

			for (int i = 0; i < videoInputs.size(); i++) 
			{
				VideoInput* prevData = nullptr;
				if (videoInputs[i].__isset.identifier)
					prevData = SdkData::GetVideoInputById(videoInputs[i].identifier);
				if (prevData == nullptr && i < SdkData::VideoInputs.size())
					prevData = &SdkData::VideoInputs[i];

				if (!videoInputs[i].__isset.identifier && prevData == nullptr)
				{
					MIXCAST_ERROR("Incomplete video input list!");
					continue;
				}

				VideoInput newData = videoInputs[i];
				TransferVideoInputData(*prevData, newData);
				newList.push_back(newData);
			}

			SdkData::VideoInputs.clear();
			for (int i = 0; i < newList.size(); i++)
				SdkData::VideoInputs.push_back(newList[i]);
		});
}
void FServiceSDKHandler::TransferVideoInputData(const VideoInput& prevData, VideoInput& newData)
{
	if (!newData.__isset.identifier && prevData.__isset.identifier) {
		newData.identifier = prevData.identifier;
		newData.__isset.identifier = true;
	}

	if (!newData.__isset.fieldOfView && prevData.__isset.fieldOfView) {
		newData.fieldOfView = prevData.fieldOfView;
		newData.__isset.fieldOfView = true;
	}
	if (!newData.__isset.currentPosition && prevData.__isset.currentPosition) {
		newData.currentPosition = prevData.currentPosition;
		newData.__isset.currentPosition = true;
	}
	if (!newData.__isset.currentRotation && prevData.__isset.currentRotation) {
		newData.currentRotation = prevData.currentRotation;
		newData.__isset.currentRotation = true;
	}

	if (!newData.__isset.projectToUser && prevData.__isset.projectToUser) {
		newData.projectToUser = prevData.projectToUser;
		newData.__isset.projectToUser = true;
	}
	if (!newData.__isset.projectToCameras && prevData.__isset.projectToCameras) {
		newData.projectToCameras = prevData.projectToCameras;
		newData.__isset.projectToCameras = true;
	}
}

void FServiceSDKHandler::UpdateViewfinderMetadata(const std::vector<Viewfinder>& viewfinders)
{
	AsyncTask(ENamedThreads::GameThread, [=]()
	{
			std::vector<Viewfinder> newList;

		for (int i = 0; i < viewfinders.size(); i++)
			{
				Viewfinder* prevData = SdkData::GetViewfinderById(viewfinders[i].identifier);

				if (!viewfinders[i].__isset.currentPosition && prevData == nullptr)
				{
					MIXCAST_ERROR("Incomplete viewfinder list!");
					continue;
				}

				Viewfinder newData = viewfinders[i];
				TransferViewfinderData(*prevData, newData);
				newList.push_back(newData);
			}

			SdkData::Viewfinders.clear();
			for (int i = 0; i < newList.size(); i++)
				SdkData::Viewfinders.push_back(newList[i]);
	});
}
void FServiceSDKHandler::TransferViewfinderData(const Viewfinder& prevData, Viewfinder& newData)
{
	if (!newData.__isset.currentPosition && prevData.__isset.currentPosition) {
		newData.currentPosition = prevData.currentPosition;
		newData.__isset.currentPosition = true;
	}
	if (!newData.__isset.currentRotation && prevData.__isset.currentRotation) {
		newData.currentRotation = prevData.currentRotation;
		newData.__isset.currentRotation = true;
	}
	if (!newData.__isset.currentScale && prevData.__isset.currentScale) {
		newData.currentScale = prevData.currentScale;
		newData.__isset.currentScale = true;
	}
}

void FServiceSDKHandler::NotifyExternalTexturesUpdated(const std::vector<SharedTexPacket>& textureInfo)
{
	//MIXCAST_LOG("TODO: implement NotifyExternalTexturesUpdated");
	AsyncTask(ENamedThreads::GameThread, [=]()
	{
		FMixCast* mod = (FMixCast*)&IMixCast::Get();
		for (const SharedTexPacket& sharedTex : textureInfo)
		{
			MIXCAST_LOG("ExternalTexturesUpdated: %s -> %d", *FString(sharedTex.id.c_str()), sharedTex.info.handle);
			mod->OnExternalTextureUpdated().Broadcast(sharedTex.id, sharedTex.info);
		}
	});
}

void FServiceSDKHandler::NotifyTrackingSpaceChanged()
{
	MIXCAST_LOG("TrackingSpaceChanged");
	AsyncTask(ENamedThreads::GameThread, [=]()
	{
		IMixCast* mod = &IMixCast::Get();
		if (mod->IsActive())
			mod->OnTrackingSpaceChanged().Broadcast();
	});

}
void FServiceSDKHandler::ResetWorldOrientation()
{
	MIXCAST_LOG("TODO: implement ResetWorldOrentiation");
}

void FServiceSDKHandler::ModifyWorldOrientation(const double degrees)
{
	MIXCAST_LOG("TODO: implement ModifyWorldOrientation");
}

void FServiceSDKHandler::SendExperienceCommand(const std::string& cmdId)
{
	FString fStr(cmdId.c_str());
	MIXCAST_LOG("SendExperienceCommand(%s)", *fStr);
	AsyncTask(ENamedThreads::GameThread, [=]()
	{
		IMixCast* mod = &IMixCast::Get();
		if (mod->IsActive())
			mod->OnCommandReceived().Broadcast(fStr);
	});
}

void FServiceSDKHandler::validate_function_list(std::vector<std::string>& _return)
{
	const int FUNC_COUNT = 15;
	const std::string FunctionList[FUNC_COUNT] = {
		"SetActivationState",
		"NotifyServiceStarted",

		"SendLegacyData",

		"UpdateTrackedObjectMetadata",
		"UpdateCameraMetadata",
		"UpdateVideoInputMetadata",
		"UpdateViewfinderMetadata",
		"UpdateDesktopMetadata",

		"NotifyExternalTexturesUpdated",

		"NotifyTrackingSpaceChanged",
		"ResetWorldOrientation",
		"ModifyWorldOrientation",

		"SendExperienceCommand",

		"validate_function_list",
		"validate_ping",
	};
	for (int i = 0; i < FUNC_COUNT; i++)
		_return.push_back(FunctionList[i]);
}

void FServiceSDKHandler::validate_ping()
{
	// Do nothing.
}
