#pragma once

#include "OwnedUnrealSharedTexture.h"
#include "OwnedUnrealSharedTexture_D3D11.h"
#include "OwnedUnrealSharedTexture_D3D12.h"
#include "Utility/ProjectUtils.hpp"

#if PLATFORM_WINDOWS
#include "Windows/AllowWindowsPlatformTypes.h"
#endif
#include <d3d11.h>
#if PLATFORM_WINDOWS
#include "Windows/HideWindowsPlatformTypes.h"
#endif

#include "CoreMinimal.h"
#include "Engine/Texture2D.h"


void* OwnedUnrealSharedTexture::GetNativeTexture(const UTexture& tex)
{
	if (!tex.IsValidLowLevel() || !tex.Resource || !tex.Resource->TextureRHI.IsValid())
		return NULL;

	return tex.Resource->TextureRHI->GetNativeResource();
}
DXGI_FORMAT OwnedUnrealSharedTexture::EnsureShareableFormat(DXGI_FORMAT format)
{
	if (format == DXGI_FORMAT_R8G8B8A8_UNORM_SRGB || format == DXGI_FORMAT_R8G8B8A8_TYPELESS)
	{
		return DXGI_FORMAT_R8G8B8A8_UNORM;
	}
	else if (format == DXGI_FORMAT_B8G8R8A8_UNORM_SRGB || format == DXGI_FORMAT_B8G8R8A8_TYPELESS)
	{
		return DXGI_FORMAT_B8G8R8A8_UNORM;
	}
	else if (format == DXGI_FORMAT_R32G32B32A32_TYPELESS)
	{
		return DXGI_FORMAT_R32G32B32A32_FLOAT;
	}
	else if (format == DXGI_FORMAT_R32_TYPELESS || format == DXGI_FORMAT_D32_FLOAT || format == DXGI_FORMAT_R32_FLOAT)
	{
		return DXGI_FORMAT_R32_FLOAT;
	}
	return format;
}

OwnedUnrealSharedTexture* OwnedUnrealSharedTexture::Create(UTexture* srcTex)
{
	switch (ProjectUtils::GetGraphicsType())
	{
	case mixcast::data::ExpGraphicsType::DX11:
		return new OwnedUnrealSharedTexture_D3D11(srcTex);
#if MIXCAST_D3D12
	case mixcast::data::ExpGraphicsType::DX12:
		return new OwnedUnrealSharedTexture_D3D12(srcTex);
#endif
	default:
		return NULL;
	}
}