/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#pragma once

#include "CoreMinimal.h"
#include "Components/ActorComponent.h"

#include "MixCastActorVisibility.generated.h"

UENUM()
namespace EMixCastCameraMode
{
	enum Type
	{
		Always,
		IfAnyConditionMet,
		IfAllConditionsMet
	};
}

UENUM()
namespace ERenderTypeCondition
{
	enum Type
	{
		NotUsed,
		Mixed,
		Virtual,
	};
}

UENUM()
namespace EBackdropStyleCondition
{
	enum Type
	{
		NotUsed,
		Virtual, 
		Physical,
	};
}

UENUM()
namespace EPerspectiveCondition
{
	enum Type
	{
		NotUsed,
		First,
		Third,
	};
}

class AExpCamera;
struct ExpFrame;

UCLASS(HideCategories = ("Actor", "LOD", "HLOD", "Input", "Cooking", "Replication", "Collision", "Tags", "Activation", "Asset User Data"), 
	ClassGroup=(MixCast), meta=(BlueprintSpawnableComponent) )
class MIXCAST_API UMixCastActorVisibility : public UActorComponent
{
	GENERATED_BODY()

public:	
	UMixCastActorVisibility();

protected:
	virtual void BeginPlay() override;
	virtual void EndPlay(const EEndPlayReason::Type EndPlayReason) override;

	void HandleFrameStarted(AExpCamera* cam);

	bool ShouldShowToCamera(const ExpFrame& frameInfo);
	bool HasAnyFlag(const ExpFrame& frameInfo);
	bool HasAllFlags(const ExpFrame& frameInfo);

private:
	FDelegateHandle FrameStartedDelegate;

public:			
	UPROPERTY(EditAnywhere, Category = "Rendering")
		bool showForRegularCameras = true;

	UPROPERTY(EditAnywhere, Category = "Rendering")
		bool showForMixCastCameras = true;
	UPROPERTY(EditAnywhere, Category = "Rendering")
		TEnumAsByte<EMixCastCameraMode::Type> mixCastCameraMode = EMixCastCameraMode::Always;
	UPROPERTY(EditAnywhere, Category = "Rendering", AdvancedDisplay)
		TEnumAsByte<ERenderTypeCondition::Type> renderTypeCondition = ERenderTypeCondition::NotUsed;
	UPROPERTY(EditAnywhere, Category = "Rendering", AdvancedDisplay)
		TEnumAsByte<EBackdropStyleCondition::Type> backdropStyleCondition = EBackdropStyleCondition::NotUsed;
	UPROPERTY(EditAnywhere, Category = "Rendering", AdvancedDisplay)
		TEnumAsByte<EPerspectiveCondition::Type> perspectiveCondition = EPerspectiveCondition::NotUsed;
};
