/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "AlphaTransfer.h"

#include "MixCastDefines.h"

#include "RenderingThread.h"
#include "SceneInterface.h"
#include "SceneView.h"
#include "Engine/TextureRenderTarget2D.h"
#include "Runtime/Renderer/Private/PostProcess/SceneRenderTargets.h"
#include "ScreenRendering.h"
#include "PipelineStateCache.h"
#include "Runtime/Renderer/Private/RendererPrivate.h"
#include "Runtime/Renderer/Private/ScenePrivate.h"
#include "Runtime/Renderer/Private/SceneRendering.h"
#include "Shader.h"
#include "GlobalShader.h"
#if UE_4_22_OR_LATER
#include "CommonRenderResources.h"
#endif

void FMixCastAlphaTransfer::TransferAlpha(
	IRendererModule& RendererModule, 
	FRHICommandListImmediate& RHICmdList, 
	FSceneRenderTargets& SceneContext, 
	FSceneView& SceneView,
	UTextureRenderTarget2D& AlphaTarget)
{
	const uint32 surfaceWidth = AlphaTarget.GetSurfaceWidth();
	const uint32 surfaceHeight = AlphaTarget.GetSurfaceHeight();

#if UE_4_23_OR_LATER
	FBox2D CopyBox(FVector2D(0.0f, 0.0f), FVector2D(surfaceWidth, surfaceHeight));
	GDynamicRHI->RHICopySubTextureRegion_RenderThread(RHICmdList, SceneContext.GetSceneColorTexture()->GetTexture2D(), AlphaTarget.GetRenderTargetResource()->GetRenderTargetTexture(), CopyBox, CopyBox);
#else
	FGraphicsPipelineStateInitializer GraphicsPSOInit;
	GraphicsPSOInit.RasterizerState = TStaticRasterizerState<FM_Solid, CM_None>::GetRHI();
	GraphicsPSOInit.DepthStencilState = TStaticDepthStencilState<false, CF_Always>::GetRHI();
	GraphicsPSOInit.BlendState = TStaticBlendState<>::GetRHI();


#if UE_4_22_OR_LATER
	FRHIRenderPassInfo RPInfo(AlphaTarget.GetRenderTargetResource()->GetRenderTargetTexture(), ERenderTargetActions::DontLoad_Store);
	RHICmdList.BeginRenderPass(RPInfo, TEXT("MixCastAlphaCapture"));
#else
	FRHIRenderTargetView TargetView(AlphaTarget.GetRenderTargetResource()->GetRenderTargetTexture(), 0, -1, ERenderTargetLoadAction::ENoAction, ERenderTargetStoreAction::EStore);
	FRHIDepthRenderTargetView DepthView;
	RHICmdList.SetRenderTargets(1, &TargetView, &DepthView, 0, NULL);
#endif

	RHICmdList.ApplyCachedRenderTargets(GraphicsPSOInit);
	RHICmdList.SetViewport(0, 0, 0, surfaceWidth, surfaceHeight, 1);

	auto ShaderMap = GetGlobalShaderMap(GMaxRHIFeatureLevel);
	TShaderMapRef<FScreenVS> VertexShader(ShaderMap);
	TShaderMapRef<FScreenPS> PixelShader(ShaderMap);

#if UE_4_22_OR_LATER
	GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = GFilterVertexDeclaration.VertexDeclarationRHI;
#else
	GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = RendererModule.GetFilterVertexDeclaration().VertexDeclarationRHI;
#endif
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = GETSAFERHISHADER_VERTEX(*VertexShader);
	GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*PixelShader);

	GraphicsPSOInit.PrimitiveType = PT_TriangleList;
	SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

	VertexShader->SetParameters(RHICmdList, SceneView.ViewUniformBuffer);
	PixelShader->SetParameters(RHICmdList, TStaticSamplerState<SF_Point>::GetRHI(), SceneContext.GetSceneColorTexture());
	
#if UE_4_19_OR_LATER
	const FViewInfo* ActualView = (FViewInfo*)(&SceneView);
#else
	const FSceneView* ActualView = &SceneView;
#endif

	RendererModule.DrawRectangle(
		RHICmdList,
		ActualView->ViewRect.Min.X, ActualView->ViewRect.Min.Y,
		ActualView->ViewRect.Width(), ActualView->ViewRect.Height(),
		ActualView->ViewRect.Min.X, ActualView->ViewRect.Min.Y,
		ActualView->ViewRect.Width(), ActualView->ViewRect.Height(),
		SceneView.UnconstrainedViewRect.Size(),
		SceneContext.GetBufferSizeXY(),
		*VertexShader,
		EDRF_UseTriangleOptimization);

#if UE_4_22_OR_LATER
	RHICmdList.EndRenderPass();
#endif
#endif
}