/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "OcclusionShader.h"
#if MIXCAST_OCCLUDE_SCREENSPACE
#include "CommonRenderResources.h"
#include "Runtime/Renderer/Private/PostProcess/SceneRenderTargets.h"

class FSimpleScreenVertexBuffer : public FVertexBuffer
{
public:
	/** Initialize the RHI for this rendering resource */
	void InitRHI()
	{
		TResourceArray<FFilterVertex, VERTEXBUFFER_ALIGNMENT> Vertices;
		Vertices.SetNumUninitialized(6);

		Vertices[0].Position = FVector4(-1, 1, 0, 1);
		Vertices[0].UV = FVector2D(0, 0);

		Vertices[1].Position = FVector4(1, 1, 0, 1);
		Vertices[1].UV = FVector2D(1, 0);

		Vertices[2].Position = FVector4(-1, -1, 0, 1);
		Vertices[2].UV = FVector2D(0, 1);

		Vertices[3].Position = FVector4(1, -1, 0, 1);
		Vertices[3].UV = FVector2D(1, 1);

		// Create vertex buffer. Fill buffer with initial data upon creation
		FRHIResourceCreateInfo CreateInfo(&Vertices);
		VertexBufferRHI = RHICreateVertexBuffer(Vertices.GetResourceDataSize(), BUF_Static, CreateInfo);
	}
};
TGlobalResource<FSimpleScreenVertexBuffer> GSimpleScreenVertexBuffer;

class FMixCastOcclusionVS : public FGlobalShader
{
	DECLARE_GLOBAL_SHADER(FMixCastOcclusionVS);

public:
	FMixCastOcclusionVS() { }
	FMixCastOcclusionVS(const ShaderMetaType::CompiledShaderInitializerType& Initializer)
		: FGlobalShader(Initializer)
	{ }

	static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
	{
		return true;
	}
};

template<int32 NumOutputs>
class FMixCastOcclusionPS : public FGlobalShader
{
public:
	DECLARE_GLOBAL_SHADER(FMixCastOcclusionPS);

	static bool ShouldCompilePermutation(const FGlobalShaderPermutationParameters& Parameters)
	{
		return IsFeatureLevelSupported(Parameters.Platform, ERHIFeatureLevel::SM5);
	}

	static void ModifyCompilationEnvironment(const FGlobalShaderPermutationParameters& Parameters, FShaderCompilerEnvironment& OutEnvironment)
	{
		FGlobalShader::ModifyCompilationEnvironment(Parameters, OutEnvironment);
		OutEnvironment.SetDefine(TEXT("NUM_OUTPUTS"), NumOutputs);
	}

protected:
#if UE_4_25_OR_LATER
	LAYOUT_FIELD(FShaderResourceParameter, DepthTexture);
	LAYOUT_FIELD(FShaderResourceParameter, DepthTextureSampler);
	LAYOUT_FIELD(FShaderParameter, DepthMultiplier);
#else
	FShaderResourceParameter DepthTexture;
	FShaderResourceParameter DepthTextureSampler;
	FShaderParameter DepthMultiplier;
#endif

public:
	FMixCastOcclusionPS() { }
	FMixCastOcclusionPS(const ShaderMetaType::CompiledShaderInitializerType& Initializer)
		: FGlobalShader(Initializer)
	{
		DepthTexture.Bind(Initializer.ParameterMap, TEXT("DepthTexture"));
		DepthTextureSampler.Bind(Initializer.ParameterMap, TEXT("DepthTextureSampler"));
		DepthMultiplier.Bind(Initializer.ParameterMap, TEXT("DepthMultiplier"));
	}

#if !UE_4_25_OR_LATER
	virtual bool Serialize(FArchive& Ar) override
	{
		bool bShaderHasOutdatedParameters = FGlobalShader::Serialize(Ar);
		Ar << DepthTexture << DepthTextureSampler << DepthMultiplier;
		return bShaderHasOutdatedParameters;
	}
#endif

	void SetParameters(FRHICommandList& RHICmdList, TShaderMapRef<FMixCastOcclusionPS<NumOutputs>> ShaderRef, FSceneView& InView, FRHITexture* DepthTextureRHI, float DepthMultiplierValue)
	{
		FRHISamplerState* SamplerStateRHI = TStaticSamplerState<SF_Bilinear>::GetRHI();
#if !UE_4_25_OR_LATER
		FRHIPixelShader* ShaderRHI = GetPixelShader();
#else
		FRHIPixelShader* ShaderRHI = ShaderRef.GetPixelShader();
#endif

		FGlobalShader::SetParameters<FViewUniformShaderParameters>(RHICmdList, ShaderRHI, InView.ViewUniformBuffer);
		SetTextureParameter(RHICmdList, ShaderRHI, DepthTexture, DepthTextureSampler, SamplerStateRHI, DepthTextureRHI);
		SetShaderValue(RHICmdList, ShaderRHI, DepthMultiplier, DepthMultiplierValue);
	}
};

IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionVS, "/Plugin/MixCast/Occlusion.usf", "MainVS", SF_Vertex);

IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<1>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<2>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<3>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<4>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<5>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<6>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<7>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);
IMPLEMENT_GLOBAL_SHADER(FMixCastOcclusionPS<8>, "/Plugin/MixCast/Occlusion.usf", "MainPS", SF_Pixel);


void FMixCastOcclusion::RenderOcclusion(FRHICommandListImmediate& RHICmdList, FSceneView& InView, FRHITexture* DepthTextureRHI, float DepthMultiplierValue)
{
	FSceneRenderTargets& SceneContext = FSceneRenderTargets::Get(RHICmdList);

	//switch to new render pass
	RHICmdList.EndRenderPass();

#if UE_4_25_OR_LATER
	FRHIRenderPassInfo RPInfo;
	int32 velIndex = -1;
#if UE_4_26_OR_LATER
	SceneContext.FillGBufferRenderPassInfo(ERenderTargetLoadAction::ELoad, RPInfo, velIndex);
#else
	int32 tangentIndex = -1;
	SceneContext.FillGBufferRenderPassInfo(ERenderTargetLoadAction::ELoad, RPInfo, velIndex, tangentIndex);
#endif
	RPInfo.DepthStencilRenderTarget.DepthStencilTarget = SceneContext.GetSceneDepthSurface();
	RPInfo.DepthStencilRenderTarget.ExclusiveDepthStencil = FExclusiveDepthStencil::DepthWrite_StencilNop;
	RPInfo.DepthStencilRenderTarget.Action = MakeDepthStencilTargetActions(ERenderTargetActions::Load_Store, ERenderTargetActions::DontLoad_DontStore);
	RHICmdList.BeginRenderPass(RPInfo, TEXT("MixCastOcclusion"));
#else
	SceneContext.BeginRenderingGBuffer(RHICmdList, ERenderTargetLoadAction::ELoad, ERenderTargetLoadAction::ELoad, FExclusiveDepthStencil::DepthWrite_StencilNop, false);
#endif


	auto ShaderMap = GetGlobalShaderMap(InView.GetFeatureLevel());

	TShaderMapRef<FMixCastOcclusionVS> VertexShader(ShaderMap);

	//prepare state
	FGraphicsPipelineStateInitializer GraphicsPSOInit;
	RHICmdList.ApplyCachedRenderTargets(GraphicsPSOInit);
	GraphicsPSOInit.BlendState = TStaticBlendState<>::GetRHI();
	GraphicsPSOInit.RasterizerState = TStaticRasterizerState<>::GetRHI();
	GraphicsPSOInit.DepthStencilState = TStaticDepthStencilState<true, CF_DepthNearOrEqual>::GetRHI();

	GraphicsPSOInit.BoundShaderState.VertexDeclarationRHI = GFilterVertexDeclaration.VertexDeclarationRHI;
	GraphicsPSOInit.PrimitiveType = PT_TriangleStrip;
#if UE_4_25_OR_LATER
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = VertexShader.GetVertexShader();
#else
	GraphicsPSOInit.BoundShaderState.VertexShaderRHI = GETSAFERHISHADER_VERTEX(*VertexShader);
#endif

	//set pixel shader
	uint32 NumActiveRenderTargets = SceneContext.GetNumGBufferTargets();
	switch (NumActiveRenderTargets)
	{
	case 5:
	{
		TShaderMapRef<FMixCastOcclusionPS<5>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif

		SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(RHICmdList, MRTPixelShader, InView, DepthTextureRHI, DepthMultiplierValue);
	}
	break;
	case 6:
	{
		TShaderMapRef<FMixCastOcclusionPS<6>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif

		SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(RHICmdList, MRTPixelShader, InView, DepthTextureRHI, DepthMultiplierValue);
	}
	break;
	default:
	case 1:
	{
		TShaderMapRef<FMixCastOcclusionPS<1>> MRTPixelShader(ShaderMap);
#if UE_4_25_OR_LATER
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = MRTPixelShader.GetPixelShader();
#else
		GraphicsPSOInit.BoundShaderState.PixelShaderRHI = GETSAFERHISHADER_PIXEL(*MRTPixelShader);
#endif

		SetGraphicsPipelineState(RHICmdList, GraphicsPSOInit);

		MRTPixelShader->SetParameters(RHICmdList, MRTPixelShader, InView, DepthTextureRHI, DepthMultiplierValue);
	}
	break;
	}

	//dispatch draw
	RHICmdList.SetStreamSource(0, GSimpleScreenVertexBuffer.VertexBufferRHI, 0);
	RHICmdList.DrawPrimitive(0, 2, 1);
}
#endif